/*
  KeePT - Plugin for KeePass 1.21
  Copyright (C) 2010 Nikolaus Hammler <niki@hammler.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/**
 * This file implements only parts which are executed in WinPT.exe's process.
 * The DLL is first loaded into WinPT.exe's address space using ExecuteRemoteThread
 * by KeePTPlugin.cpp
 *
 * The DllMain calls InjectPT() but only in case the DLL is loaded into WinPT.exe
 * (not KeePass.exe!!)
 * This procedure installs a hook using apihijack which overwrites the DialogBoxParamA
 * API function.
 *
 * In case WinPT.exe calls DialogBoxParamA API function to display a password dialog,
 * the password is instead retrieved by KeePT plugin by sending a window message and
 * passing the password via shared memory
 */

// TODO: Make a proxy for _MVC functions
#define _CRT_SECURE_NO_WARNINGS

#include <windows.h>
#include <shlwapi.h>
#include <assert.h>
#include "KeePTPlugin.h"
#include "WinPTHooks.h"
#include "WinPTExec.h"
#include "WinPT/src/resource.h"

// These are the shared memory variables which are shared among
// all instances of KeePT.dll
// Defined in KeePTPlugin.dll as dataseg .SHARED
extern HWND g_shhKeeWnd;
// END memory shared variables

static DLGPROC OrigPasswordDlgProc = 0;

// that API function DialogBoxParamA in user32.dll shall be overwritten
// by MyDialogBoxParam
static SDLLHook WPTHook = 
{
    "USER32.DLL",
    false, NULL,    // Default hook disabled, NULL function pointer.
    {
        { "DialogBoxParamA", (void*)MyDialogBoxParam },
        { NULL, NULL }
    }
};

/**
 * This is the overwritten function of the original password dialog callback
 * We overwrite it in order to allocate memory for the password: We auto-fill
 * the password control with MAXPWDLEN-1 spaces in WM_INITDIALOG
 *
 * After returning from DialogBoxParam, cds->pwd contains a string of
 * MAXPWDLEN-1 spaces which will be overwritten by the "real" password
 */
INT_PTR CALLBACK MyPasswordDlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	assert( OrigPasswordDlgProc != NULL );

	if(msg != WM_INITDIALOG)
		return OrigPasswordDlgProc(hWnd, msg, wParam, lParam);

	assert( g_shhKeeWnd != NULL );

	struct passphrase_cb_s *c = (struct passphrase_cb_s*)lParam;

	assert( c != NULL );

	char *fooData = new char[MAXPWDLEN];
	memset(fooData, ' ', MAXPWDLEN - 1);
	fooData[MAXPWDLEN - 1] = '\0';

	INT_PTR ret = OrigPasswordDlgProc(hWnd, msg, wParam, lParam);
	SetDlgItemText(hWnd, c->gpg_cmd == GPG_CMD_DECRYPT ? IDC_DECRYPT_PWD : IDC_DECRYPT_SIGN_PWD, fooData);
	delete[] fooData;

	SendMessage(hWnd, WM_COMMAND, MAKEWPARAM(IDOK, BN_CLICKED), 0);
	return ret;
}

bool QuerySecurePassword(char *keyid, char *dst)
{
	bool ret = false;
	char mapname[5 + 8 + 1];
	PSID usersid = NULL;
	PSECURITY_DESCRIPTOR psd = NULL;
	HANDLE filemap = NULL;
	char *p = NULL;

	if(!g_shhKeeWnd)
		return false;

	wsprintf(mapname, "KeePT%08x", (unsigned)GetCurrentThreadId());
	
	usersid = GetMySID();
	if(!usersid)
		goto cleanup;

	psd = (PSECURITY_DESCRIPTOR)LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH);
	if(!psd)
		goto cleanup;

	if(InitializeSecurityDescriptor(psd, SECURITY_DESCRIPTOR_REVISION) == 0)
		goto cleanup;

	if(SetSecurityDescriptorOwner(psd, usersid, FALSE) == 0)
		goto cleanup;

	SECURITY_ATTRIBUTES sa;
	sa.nLength = sizeof(sa);
	sa.bInheritHandle = FALSE;
	sa.lpSecurityDescriptor = psd;

	filemap = CreateFileMapping(INVALID_HANDLE_VALUE, &sa, PAGE_READWRITE, 0, MAXPWDLEN + WINPT_KEYID_MAXLEN, mapname);
	if(filemap == NULL || filemap == INVALID_HANDLE_VALUE)
		goto cleanup;

	p = (char*)MapViewOfFile(filemap, FILE_MAP_WRITE, 0, 0, 0);
	if(!p)
		goto cleanup;

	if(VirtualLock(p, MAXPWDLEN + WINPT_KEYID_MAXLEN) == 0)
		goto cleanup;

	strncpy(p, keyid, WINPT_KEYID_MAXLEN);

	COPYDATASTRUCT cds;
	cds.dwData = WM_GETPTPWD_MAGIC;
	cds.lpData = mapname;
	cds.cbData = 1 + (int)strlen(mapname);

	if(SendMessage(g_shhKeeWnd, WM_COPYDATA, (WPARAM)NULL, (LPARAM)&cds) == WM_GETPTPWD_MAGIC)
	{
		strncpy(dst, p, MAXPWDLEN);
		ret = true;
	}

	memset(p, 0x00, MAXPWDLEN + WINPT_KEYID_MAXLEN);

	VirtualUnlock(p, MAXPWDLEN + WINPT_KEYID_MAXLEN);

	goto cleanup;

cleanup:
	if(usersid)
		free(usersid);
	if(psd)
		LocalFree(psd);
	if(p)
		UnmapViewOfFile(p);

	if(filemap)
		CloseHandle(filemap);

	return ret;
}

/**
 * This function is called everytime when WinPT.exe displays a dialog
 */
int WINAPI MyDialogBoxParam(HINSTANCE hInstance, LPCTSTR lpTemplateName, HWND hWndParent, DLGPROC lpDialogFunc, LPARAM dwInitParam)
{
#ifdef _MSC_VER
#pragma warning(disable:4311)
#endif
	DWORD id = (DWORD)lpTemplateName;
#ifdef _MSC_VER
#pragma warning(default:4311)
#endif

	DialogBoxParam_Type OrigDialogBoxParam = (DialogBoxParam_Type)WPTHook.Functions[0].OrigFn;
	assert( OrigDialogBoxParam != NULL );

	switch(id)
	{
		case IDD_WINPT_DECRYPT:
		case IDD_WINPT_DECRYPT_SIGN:
			break;
		default:
			goto def;
	}

	if(!g_shhKeeWnd)
		goto def;

	struct passphrase_cb_s *c = (struct passphrase_cb_s*)dwInitParam;
	if(!c)
		goto def;

	// This is only done in order to allocate MAXPWDLEN bytes in
	// cds->pwd
	// At first, I just allocated memory with new/malloc
	// However, this works only as long as the CRT in WinPT (for deallocating the memory)
	// matches KeePT's CRT. If not, WinPT is likely to crash
	// Therefore we need to let WinPT allocate the memory for the passphrase for us...
	OrigPasswordDlgProc = lpDialogFunc;
	if(OrigDialogBoxParam(hInstance, lpTemplateName, hWndParent, MyPasswordDlgProc, dwInitParam) != TRUE)
		goto def;
	if(!c->pwd)
		goto def;

	// check if allocation was successful ...
	int lenIs = (int)strnlen(c->pwd, MAXPWDLEN);
	if(lenIs != MAXPWDLEN - 1)
		goto def;

	if(QuerySecurePassword(c->keyid, c->pwd))
	{
		return TRUE;
	}

def:
	return OrigDialogBoxParam(hInstance, lpTemplateName, hWndParent, lpDialogFunc, dwInitParam);
}

bool InjectPT(HINSTANCE hModule)
{
	char modulename[MAX_PATH];
	memset(modulename, 0, MAX_PATH);
	DisableThreadLibraryCalls(hModule);
	GetModuleFileName(GetModuleHandle(NULL), modulename, MAX_PATH-1);
	PathStripPath(modulename);
#ifdef _MSC_VER
	// this is a POSIX function which is deprecated starting with VS2005
	// _stricmp is the ISO compliant name
	if(_stricmp(modulename, "WinPT.exe") == 0)
#else
	if(stricmp(modulename, "WinPT.exe") == 0)
#endif
	{
		return HookAPICalls(&WPTHook);
	}
	return true;
}





